﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using Microsoft.Phone.Controls;
using System.Net.Sockets;
using System.Text;

namespace MulticastDemo
{
    public partial class MainPage : PhoneApplicationPage
    {
        // The address of the multicast group to join.
        // Must be in the range from 224.0.0.0 to 239.255.255.255
        private const string GROUP_ADDRESS = "224.0.1.1";

        // The port over which to communicate to the multicast group
        private const int GROUP_PORT = 52274;

        // A client receiver for multicast traffic from any source
        UdpAnySourceMulticastClient _client = null;

        // true if we have joined the multicast group; otherwise, false
        bool _joined = false;

        // Buffer for incoming data
        private byte[] _receiveBuffer;

        // Maximum size of a message in this communication
        private const int MAX_MESSAGE_SIZE = 512;


        // Constructor
        public MainPage()
        {
            InitializeComponent();
        }

        private void btnSend_Click(object sender, RoutedEventArgs e)
        {
            // Don't send empty messages.
            if (!String.IsNullOrWhiteSpace(txtInput.Text))
            {
                //Send whatever the user has typed into the input textbox to the multicast group.
                Send(txtInput.Text);
            }
        }


        /// <summary>
        /// Log text to the lbLog ListBox.
        /// </summary>
        /// <param name="message">The message to write to the lbLog ListBox</param>
        /// <param name="isOutgoing">True if the message is an outgoing (client to server) message; otherwise, False</param>
        /// <remarks>We differentiate between a message from the client and server 
        /// by prepending each line  with ">>" and "<<" respectively.</remarks>
        private void Log(string message, bool isOutgoing)
        {
            if (string.IsNullOrWhiteSpace(message.Trim('\0')))
                return;

            // Always make sure to do this on the UI thread.
            Deployment.Current.Dispatcher.BeginInvoke(
            () =>
            {
                string direction = (isOutgoing) ? ">> " : "<< ";
                string timestamp = DateTime.Now.ToString("HH:mm:ss");
                message = timestamp + direction + message;
                lbLog.Items.Add(message);

                // Make sure that the item we added is visible to the user.
                lbLog.ScrollIntoView(message);
            });

        }

        /// <summary>
        /// Override the OnNavigatedTo method.
        /// </summary>
        protected override void OnNavigatedTo(System.Windows.Navigation.NavigationEventArgs e)
        {
            base.OnNavigatedTo(e);

            // Create a new UdpAnySourceMulticastClient instance and join the group.
            Join();
        }

        /// <summary>
        /// Create a new UdpAnySourceMulticastClient instance and join the group.
        /// </summary>
        private void Join()
        {
            // Initialize the receive buffer
            _receiveBuffer = new byte[MAX_MESSAGE_SIZE];

            // Create the UdpAnySourceMulticastClient instance using the defined 
            // GROUP_ADDRESS and GROUP_PORT constants. UdpAnySourceMulticastClient is a 
            // client receiver for multicast traffic from any source, also known as Any Source Multicast (ASM)
            _client = new UdpAnySourceMulticastClient(IPAddress.Parse(GROUP_ADDRESS), GROUP_PORT);

            // Make a request to join the group.
            _client.BeginJoinGroup(
                result =>
                {
                    // Complete the join
                    _client.EndJoinGroup(result);

                    // The MulticastLoopback property controls whether you receive multicast 
                    // packets that you send to the multicast group. Default value is true, 
                    // meaning that you also receive the packets you send to the multicast group. 
                    // To stop receiving these packets, you can set the property following to false
                    _client.MulticastLoopback = true;

                    // Set a flag indicating that we have now joined the multicast group 
                    _joined = true;

                    // Let others know we have joined by sending out a message to the group.
                    Send("Joined the group");

                    // Wait for data from the group. This is an asynchronous operation 
                    // and will not block the UI thread.
                    Receive();
                }, null);
        }

        /// <summary>
        /// Send the given message to the multicast group.
        /// </summary>
        /// <param name="message">The message to send</param>
        private void Send(string message)
        {
            // Attempt the send only if you have already joined the group.
            if (_joined)
            {
                byte[] data = Encoding.UTF8.GetBytes(message);
                _client.BeginSendToGroup(data, 0, data.Length,
                    result =>
                    {
                        _client.EndSendToGroup(result);

                        // Log what we just sent
                        Log(message, true);

                    }, null);
            }
            else
            {
                Log("Message was not sent since you are not joined to the group", true);
            }
        }

        /// <summary>
        /// Receive data from the group and log it.
        /// </summary>
        private void Receive()
        {
            // Only attempt the send if you have already joined the group
            if (_joined)
            {
                Array.Clear(_receiveBuffer, 0, _receiveBuffer.Length);
                _client.BeginReceiveFromGroup(_receiveBuffer, 0, _receiveBuffer.Length,
                    result =>
                    {
                        IPEndPoint source;

                        // Complete the asynchronous operation. The source field will 
                        // contain the IP address of the device that sent the message
                        _client.EndReceiveFromGroup(result, out source);

                        // Get the received data from the buffer.
                        string dataReceived = Encoding.UTF8.GetString(_receiveBuffer, 0, _receiveBuffer.Length);

                        // Create a log entry.
                        string message = String.Format("[{0}]: {1}", source.Address.ToString(), dataReceived);

                        // Log it.
                        Log(message, false);

                        // Call receive again to continue to "listen" for the next message from the group
                        Receive();
                    }, null);
            }
            else
            {
                Log("Cannot receive. You are currently not joined to the group", true);
            }
        }

    }
}